@@ -1,5 +1,6 @@
 from zope.interface import implements, Interface
 from zope.component import adapts
+from zope.component import queryAdapter
 from zope.annotation.interfaces import IAnnotations
 
 from plone.portlets.interfaces import IPortletContext
@@ -31,23 +32,33 @@ class PortletRetriever(object):
         """Work out which portlets to display, returning a list of dicts
         describing assignments to render.
         """
-        pcontext = IPortletContext(self.context, None)
+
+        if IPortletContext.providedBy(self.context):
+            pcontext = self.context
+        else:
+            pcontext = queryAdapter(self.context, IPortletContext)
+
         if pcontext is None:
             return []
             
         # Holds a list of (category, key, assignment).
         categories = [] 
+        
+        # Keeps track of the blacklisting status for global categores 
+        # (user, group, content type). The status is either True (blocked)
+        # or False (not blocked).
         blacklisted = {}
+                
+        # This is the name of the manager (column) we're rendering
         manager = self.storage.__name__
         
-        # Fetch blacklisting status for each category
+        # 1. Fetch blacklisting status for each global category
 
         # First, find out which categories we will need to determine
         # blacklist status for
         
         for category, key in pcontext.globalPortletCategories(False):
             blacklisted[category] = None
-        blacklisted[CONTEXT_CATEGORY] = None
         
         # Then walk the content hierarchy to find out what blacklist status
         # was assigned. Note that the blacklist is tri-state; if it's None it
@@ -57,75 +68,59 @@ class PortletRetriever(object):
         # value for a given item will set the appropriate value. Parents of
         # this item that also set a black- or white-list value will then be
         # ignored.
-        
+
+        # Whilst walking the hierarchy, we also collect parent portlets, 
+        # until we hit the first block.
+
         current = self.context
         currentpc = pcontext
         blacklistFetched = set()
+        parentsBlocked = False
         
         while current is not None and currentpc is not None:
-            assignable = ILocalPortletAssignable(current, None)
+            if ILocalPortletAssignable.providedBy(current):
+                assignable = current
+            else:
+                assignable = queryAdapter(current, ILocalPortletAssignable)
+
             if assignable is not None:
-                annotations = IAnnotations(assignable)
+                if IAnnotations.providedBy(assignable):
+                    annotations = assignable
+                else:
+                    annotations = queryAdapter(assignable, IAnnotations)
+
+                if not parentsBlocked:
+                    local = annotations.get(CONTEXT_ASSIGNMENT_KEY, None)
+                    if local is not None:
+                        localManager = local.get(manager, None)
+                        if localManager is not None:
+                            categories.extend([(CONTEXT_CATEGORY, currentpc.uid, a) for a in localManager.values()])
+
                 blacklistStatus = annotations.get(CONTEXT_BLACKLIST_STATUS_KEY, {}).get(manager, None)
                 if blacklistStatus is not None:
                     for cat, status in blacklistStatus.items():
-                        if blacklisted.get(cat, False) is None:
-                            blacklisted[cat] = status
-                        if status is not None:
-                            blacklistFetched.add(cat)
+                        if cat == CONTEXT_CATEGORY:
+                            if not parentsBlocked and status == True:
+                                parentsBlocked = True
+                        else: # global portlet categories
+                            if blacklisted.get(cat, False) is None:
+                                blacklisted[cat] = status
+                            if status is not None:
+                                blacklistFetched.add(cat)
+                                
+            # We can abort if parents are blocked and we've fetched all
+            # blacklist statuses
             
-            # If we have found "real" settings (i.e. not "acquire") for all 
-            # the different blacklisting types we need, stop
-            if len(blacklistFetched) >= len(blacklisted):
+            if parentsBlocked and len(blacklistFetched) == len(blacklisted):
                 break
-                
+                    
             # Check the parent - if there is no parent, we will stop
             current = currentpc.getParent()
             if current is not None:
-                currentpc = IPortletContext(current, None)
-        
-        # Get contextual portlets and black/white listings if possible. 
-        # Note that not every parent may be an ILocalPortletAssignable, in 
-        # which case we simply skip past this parent.
-        
-        current = self.context
-        currentpc = pcontext
-        checkingContext = True
-        
-        while current is not None and currentpc is not None:
-            assignable = ILocalPortletAssignable(current, None)
-            if assignable is not None:
-                annotations = IAnnotations(assignable)
-                
-                local = annotations.get(CONTEXT_ASSIGNMENT_KEY, None)
-                if local is not None:
-                    localManager = local.get(manager, None)
-                    if localManager is not None:
-                        categories.extend([(CONTEXT_CATEGORY, currentpc.uid, a) for a in localManager.values()])
-
-                # Abort after fetching the current context's items if parent
-                # portlets are blacklisted. This setting may have been
-                # acquired from a parent, but we still apply the setting
-                # relative to the context!
-                
-                if blacklisted[CONTEXT_CATEGORY]:
-                    break
-                
-                # Even if we are actually fetching some parent items, we
-                # still want to stop collecting portlets if we hit a parent
-                # that itself blacklisted its parents.
-                
-                if checkingContext:
-                    checkingContext = False
+                if IPortletContext.providedBy(current):
+                    currentpc = current
                 else:
-                    blacklistStatus = annotations.get(CONTEXT_BLACKLIST_STATUS_KEY, {}).get(manager, {})
-                    if blacklistStatus.get(CONTEXT_CATEGORY, None):
-                        break
-
-            # Check the parent - if there is no parent, we will stop
-            current = currentpc.getParent()
-            if current is not None:
-                currentpc = IPortletContext(current, None)
+                    currentpc = queryAdapter(current, IPortletContext)
         
         # Get all global mappings for non-blacklisted categories
         
@@ -159,10 +154,14 @@ class PlacelessPortletRetriever(PortletRetriever):
         self.storage = storage
         
     def getPortlets(self):
-        pcontext = IPortletContext(self.context, None)
+        if IPortletContext.providedBy(self.context):
+            pcontext = self.context
+        else:
+            pcontext = queryAdapter(self.context, IPortletContext)
+
         if pcontext is None:
             return []
-            
+
         assignments = []
         for category, key in pcontext.globalPortletCategories(True):
             mapping = self.storage.get(category, None)
